@@ -75,6 +75,20 @@
 #   endif
 #endif
 
+#ifndef HAVE_CLEARENV
+int clearenv (void)
+{
+	/* According to:
+	 *   http://hg.dovecot.org/dovecot-2.0/file/74d9f61e224d/src/lib/env-util.c#l56
+	 * simply setting "environ = NULL" crashes some systems.  Creating a
+	 * new environment consisting of just a terminator is indeed
+	 * probably the best we can do.
+	 */
+	environ = XCALLOC (1, char *);
+	return 0;
+}
+#endif
+
 /* ---------------------------------------------------------------------- */
 
 /* Functions to build individual commands. */
@@ -255,6 +269,7 @@ pipecmd *pipecmd_new_argstr (const char *argstr)
 		 * work when being executed directly. To work around this,
 		 * we just drop "exec" if it appears at the start of argstr.
 		 */
+		free (arg);
 		arg = argstr_get_word (&argstr);
 		if (!arg)
 			error (FATAL, 0,
@@ -733,6 +748,11 @@ void pipecmd_exec (pipecmd *cmd)
 			struct pipecmd_sequence *cmds = &cmd->u.sequence;
 			struct sigaction sa;
 
+			/* Flush all pending output so that subprocesses
+			 * don't inherit it.
+			 */
+			fflush (NULL);
+
 			/* pipeline_start will have blocked SIGCHLD. We like
 			 * it that way. Lose the parent's signal handler,
 			 * though.
@@ -1413,7 +1433,7 @@ void pipeline_start (pipeline *p)
 				p->outfd = pdes[0];
 			output_read = pdes[0];
 			output_write = pdes[1];
-		} else if (i == p->ncommands - 1) {
+		} else {
 			if (p->redirect_out == REDIRECT_FD)
 				output_write = p->want_out;
 			else if (p->redirect_out == REDIRECT_FILE_NAME) {
@@ -1660,6 +1680,21 @@ int pipeline_wait_all (pipeline *p, int **statuses, int *n_statuses)
 		if (active_pipelines[i] == p)
 			active_pipelines[i] = NULL;
 
+	/* It isn't normally safe to compactify active_pipelines as we go,
+	 * because it's used by a signal handler.  However, if it's entirely
+	 * empty, we can safely clean it up now.  This prevents the table
+	 * growing without bound, not to mention pacifying valgrind.
+	 */
+	for (i = 0; i < n_active_pipelines; ++i)
+		if (active_pipelines[i])
+			break;
+	if (i == n_active_pipelines) {
+		n_active_pipelines = 0;
+		max_active_pipelines = 0;
+		free (active_pipelines);
+		active_pipelines = NULL;
+	}
+
 	if (statuses && n_statuses) {
 		*statuses = xnmalloc (p->ncommands, sizeof **statuses);
 		*n_statuses = p->ncommands;
@@ -2147,7 +2182,7 @@ static const char *get_line (pipeline *p, size_t *outlen)
 		if (!buffer || plen == 0)
 			return NULL;
 
-		end = memchr (buffer + block * i, '\n', plen);
+		end = memchr (buffer + block * i, '\n', plen - block * i);
 		if (!end && plen < block * (i + 1))
 			/* end of file, no newline found */
 			end = buffer + plen - 1;
